# On processing the contents of a module body

Author: Óscar Martín
Last modified: May 2021



For reasons that I have forgotten, I decided that syntax objects exist *out of a module*. That is, an equation, for example, is not tied to a particular module. An equation is just a lhs (which, in turn, is a string representing a term), a rhs and maybe a condition, and all of them exist by themselves. This is only possible from a first-pass perspective, that is, because we treat terms as just strings. Also, a sort declaration exists with no module being necessary. Thus, a `ModBody` is just a list of instructions, that we make into a class so that we can define methods on it, but that must not be thought of as tied to a module. Indeed, it does not know the name of the module it belongs to, because it belongs to no module.

The class `Body` only is able to return the whole list of its instructions. It is frequently the case that we need to access some particular type of instructions, like all rules, or all imports. This is the aim of `BodyProcessor`. It takes as parameter a body, and adds to it a series of methods (and properties, in the Python sense) to access particular elements.

As soon as the `BodyProcessor` object is created, it explores the set of instructions in the body and computes all sets of instructions by type. This may seem an unnecessary work, because maybe not all of these sets are going to be needed. But it is not. Each time we create a `BodyProcessor` object, we need to access some set of instructions, and to compute such set, we need to go through all the instructions, given that they are not in any particular order. I am taking advantage of this traversal of the whole body to process it all in one go.

A few methods receive a `BodyProcessor`, instead of a `Body`, so as to make as few computations as possible. That is, if they received a `Body`, they would need to create the `BodyProcessor` themselves, replicating work.

This may be a good place to discuss a little technical detail that appears in several places, including some methods in this class `BodyProcessor`. The issue shows up When a method must return the concatenation of several list previously computed. For example, the method `all_imports` returns the concatenation of `protectings`, `extendings` and `includings`. If I return it just like this:
```maude
return self.protectings + self.extendings + self.includings
```
Pycharm's linter complains about the types of the three operands, because, indeed, each is a list of different things. However, if I do it in three steps:
```maude
res = self.protectings
res += self.extendings
res += self.includings
return res
```
it stops complaining. I do not understand why this is different from the previous code. I prefer that Pycharm does not complain, so that I can trust something noteworthy is happening when the green checkmark does not appear. But there is still a problem with the previous code: the first instruction `res = self.protectings` does not create a new list `res`, but a new name `res` for the same list `self.protectings`. So, when I concatenate the other two, they are all being concatenated to `self.protectings`. There are two ways to solve this. Either say explicitly that we want a copy:
```maude
res = self.protectings.copy()
```
or the following, that I have preferred in my code:
```maude
res = []
res += self.protectings
res += self.extendings
res += self.includings
return res
```
