# Lark and `metaParse`---a tale of two parsers

Author: Óscar Martín
Last modified: Apr 2021



The implementation of cMaude uses two parsing tools: Lark (see [here](https://github.com/lark-parser/lark)) and Maude's `metaParse` (see Section 17.6.13 of the manual for Maude version 3.1). This document is an exploration of each one's capabilities and why I use both.

Maude parsing tools are is very well suited for parsing Maude code. This is so in at least two respects: tokenization and mixfix operators. What constitutes a lexical token in Maude is not what I am used to in other languages. Any character is allowed to be or form part of a token (of an identifier, for example), with the exception of the seven breaking characters, `()[]{},`, that are tokens by themselves but cannot be part of larger ones (except if escaped with a backquote). Thus, `ab(cd.ed;fg hi` is a sequence of four tokens: `ab`, `(`, `cd.ed;fg`, and `hi`. In Lark, terminals are described by regexes, and it has taken some work and ingenuity to get the right regex for the task. Maude has a function `tokenize`, declared in the prelude (as a hook) but implemented in C++, that does it at once.

Something similar happens with operators with underscores in their names to represent placeholders for arguments. Very particularly, the use of empty syntax operators, declared as `op __ : S T -> U .` is difficult to parse. In the presence of such declaration, the method `parseTerm` readily parses `s t` as `'__['s.S, 't.T]`. But, because there is no syntax, it may be difficult for other parsers like Lark. When we want to perform name qualification, as I discuss in another document, Lark is of little use in the presence of empty syntax.

I must mention Rubén Rubio's Python bindings for Maude, see [here](https://fadoss.github.io/maude-bindings/). It allows running Maude functions from Python. It provides a library of classes and methods to input code into Maude interpreter, perform reductions, etc. One of the methods provided is called `parseTerm`. It is a method in the class `Module`. It receives a string supposedly containing a term in the module. It returns the object `Term` resulting from the parsing. So it is not even needed to use first `tokenize` and then `metaParse`. As we discuss below, this is not always enough, and sometimes we need to work at the metalevel using these bindings, which makes things a little more complicated.

So, why did I decide to use Lark, instead of Maude tools?

There is nothing in particular about Lark (though it works quite well). I wanted to use a Python based parser, for the same reason that I wanted to use Python instead of following the usual path of extending Full Maude: to explore a new way. Also... because Python is nice to work with, and extending Maude is not always a joyful task.

It must be said, as well, that I didn't foresee the need to parse into terms for qualification. I looked for a parser based on BNF, so that the resulting grammars would be as standard and easy to understand as possible. Describing Maude syntax in Lark's BNF (rather, EBNF) flavour is not difficult task, appart of the terminals, as discussed above. What is difficult is parsing inside terms, which I never intended to do. Then, the need for a careful qualification inside terms came, and the need for a use of Maude's parsing capabilities.

There is an additional advantage in using Maude's parsing capabilities: Full Maude has coded all the syntax of Maude, that I can use as well.

So, then, why not using Maude's functions for both parsing passes? In part, because I had already devoted a good amount of time to the writing of (part of) Maude's grammar in Lark's BNF. But there are other reasons.

One such reason is that BNF is a standard format. The fact that Chapter 24 in the Maude manual describes Maude's grammar in BNF speaks by itself. (Maude's C++ implementation uses lex and yacc, by the way.)

The fact that Maude's tools are so focused in parsing Maude or Maude-like syntax may be a drawback for some extensions. For example, it may be difficult to parse languages where identifiers are limited to letters, numbers and underscores. I have no plans to use my tool to parse anything but simple extensions of Maude, but... who knows. Maybe translating from some other specification language into Maude?

A point worth mentioning is that parsing using Maude tools requires working at the metalevel, which makes things more complicated. I'll try to explain why. Suppose a module's signature contains
```maude
sort S .
op f_ : S -> S .
op a : -> S .
```
This gives enough information to parse the term `f f a`, because all elements in this term have been declared. The situation is different when arbitrary tokens must be accepted, as when declaring the syntax for a sort declaration: `op sort_. : ...` The argument can be any token, that is, basically, any sequence of characters (with some requirements). But this is not something that we can define using Maude. The solution is working at the metalevel, where every object is represented as a `Qid` (yes, that's an oversimplification). Full Maude declares a sort `@Token@` and some specialized ones like `@SortToken@`, `@ViewToken@`, etc. They are all metalevel entities. Moreover, they are declared using hooks, because, at the end, they are defined in Maude's C++ implementation.

Working at the metalevel entails using `tokenize` and `metaParse`. The way to use them from Python is, as discussed above, using Rubén's bindings. But they do not provide equivalents to those functions. The way to go, then, is using `parseTerm` to parse and reduce metalevel expressions. For example:
```py
mod = maude.getModule('LEXICAL')
term = mod.parseTerm(f'tokenize("sort S .")')
term.reduce()
```
After executing this, `term` contains a term with three `Qid`s: `'sort`, `'S`, and `'.`. Something similar but a little more involved is used with `metaParse` instead of `tokenize`. 

All this makes using Maude tools for parsing complicated, though, indeed, all the complexities can be hidden inside convenient functions or even classes. There is another document in this collection explaining the way.

(As a side note, Rubén's tool does not include the concept of session: all communication from Python to Maude is in a single session. Thus, if several `Qualifier` objects are used, their lifespans cannot overlap, or unexpected behaviours would happen. As Rubén says in the documentation, the package `multiprocessing` can help with that.)
