# On qualifying names

Author: Óscar Martín
Last modified: Apr 2021



The key part of our implementation consists in combining several system modules into a single one that represents their composition. The rules of the resulting module are obtained by composing the components' rules; the rest is, essentially, a disjoint union of the contents of the components: declarations, equations, memberships. The implementation needs to ensure such disjointness. An easy solution, specially for a prototype, is to trust the users and assume they have not used same names in different components, and maybe warn them if they do. This is the approach taken by Maude's interpreter when importing modules. But, in our theoretical developements, we have insisted that the user must be able to code each component with no regard to others, even not knowing whether it will be later composed. So we must provide the tools to make it possible.

What we do is to *qualify* each sort, operator, and variable name in each module before joining them. Namely, we change `name` to `<MODNAME|name>`, where `MODNAME` is the name of the module on which the sort, operator or variable was originally declared. This ensures disjointness. But the process is full of small important details.

To begin with, we always perform qualification on a *semi-flattened* module. *Flattening* here is in the sense of turning imported modules into actual code inserted verbatim into the importing module. We perform flattening for any system module that is imported, and also for any functional module that is imported in `including` mode, but we leave alone importations of functional modules in `protecting` or `extending` modes. Thus, a module like
```maude
mod KK is
   pr S1 . ex S2 . inc S3 .
   pr F1 . ex F2 . inc F3 .
   < other instructions >
endm
```
where `S1`, `S2`, `S3` are system modules and `F1`, `F2`, `F3` are functional ones, is semi-flatenned to
```maude
mod KK is
   pr F1 . ex F2 .
   < verbatim code from S1 >
   < verbatim code from S2 >
   < verbatim code from S3 >
   < verbatim code from F3 >
   < other instructions >
endm
```
That's what we mean by *semi-flattening*. The reason for this is explained in the text of my PhD thesis. (Anyway, it can be easily justified that, when obtaining a composed module, imported modules like `BOOL` and `NAT` must be shared, and not flattened and then qualified.) So, first, we semi-flatten; then, we qualify. This, indeed, is not important to the implementation of the qualification procedure, but only to the way we use it.

Anyway, the method we use to parse terms, which is explained below (method `parseTerm` in Rubén's bindings) does not look inside imported modules, only takes into account the signature present in the module. So some form of flattening may be needed.

The previous paragraphs show what we do; how we do it is a different matter. The problem is that we must perform qualification inside terms. Maude allows the declaration and use of operators with a very flexible *mixfix* syntax. A consequence is that parsing in Maude needs to be a two-pass process. I discuss this at more length in another document in this collection. The first pass is only based on Maude's syntax and, thus, cannot parse inside terms, whose grammatical structure is determined by the declarations of operators (and sorts and subsorts) in each particular module. In this first pass, terms are just identified and delimited (turned into *bubbles*). Then, based on each module's *signature* (that is, declarations, as already parsed), the second pass solves bubbles, and parse trees are produced for their terms.

Usually, our implementation does not need the second pass, there is no need to go into terms. For instance, two rules $t_1 \rightarrow t'_1$ and $t_2 \rightarrow t'_2$ are combined into $\langle t_1, t_2\rangle \rightarrow \langle t'_1, t'_2\rangle$, whatever those four terms are. But, for qualifying, we do need the second pass so that identifiers in a term get renamed.

For the first pass we use [Lark](https://github.com/lark-parser/lark), a package for parsing in Python. Grammar definitions for Lark are written in a BNF-like syntax. This makes it unsuitable for the second pass, because, in that case, the input grammar is given by the module's signature, as written in Maude's syntax. It is certainly possible to translate Maude declarations into BNF form... to some extent. Some features of Maude's are not easily translated into BNF (if at all possible). Particularly difficult to translate are attributes like `assoc`, and `id:`, and also operators with underscores in their names, especially empty syntax operators `__`. Also, even if we manage to get the parse tree for the qualified term, prettyprinting it correctly is not a trivial task.

So, for the second pass, I use Maude's parsing capabilities, that we can invoke from Python thanks to [Rubén Rubio's bindings](https://github.com/fadoss/maude-bindings). If it seems odd that we use two parsers... well, it probably is. There is in this collection a document on that.

Let us point out that, strictly speaking, we do not need to parse terms completely, we do not need a parse tree. We only need to delimit, inside each term-bubble, the identifiers that need to be qualified (which can be easily drawn from the declarations). It is not difficult to find, within a string, where an identifier starts and ends, based on the set of characters allowed in identifiers. In principle, a substring replacement on those substrings would do. We tried this approach, and it worked rather well, but with some problems. (I have saved the code I used for this somewhere.) The case of empty-syntax operators is easy to understand: because there is no syntax, there is nothing to find and replace. But any operator id with underscores is problematic. And also, ids with backquotes to represent whitespaces. In short, this method works rather well if we only use *standard* names for operators, meaning no use of underscores. Inlined variables, that is, the ones that are not declared, but used on the fly with the syntax `V:Sort`, are another source of problems, precisely because they are not declared. We decided that the effort was worth to save Maude's mixfix, user-defined syntax.

As any other tool, Rubén's has its limitations. What we would like to be able to code is a function to which we can hand a Maude module containing a signature (a grammar) and a string containing a term in that signature (a bubble) and receive as the result a string containing the qualified term. The way to do it is to parse the bubble in the signature to obtain a term tree, qualify identifiers in it to obtain a qualified term tree, and prettyprint the latter. Terms, however, exist on a signature, so the qualified term tree cannot be built unless its signature module is produced first. This is a problem, because Rubén's tool allows the construction of terms (on given modules) but not objects of other syntactic categories---in particular, no construction of modules. From some point of view, this may not be too bad, because the construction of modules containing only signatures (and, thus, no terms) is a first-pass task. Anyway, it seems there is no other solution than to work with both the original signature and the qualified one, and to translate a term from the first to the second. (Indeed, the construction of modules *is* possible working at the metalevel and then going down, but I have not explored this way.)

At some point I implemented all this as a *translator*, that is, a tool (a class, indeed) which is given two languages (in the form of two signatures), and a dictionary of correspondences, and performs the translation. This worked quite well, but had some problems. One is the treatment of inlined variables, which are not part of the signature and, thus, cannot be included in the dictionary. The other is that, well, we do not need a general translator, but a very particular case of it. Using a translator from the main procedure entails the creation of the qualified version of the signature and the dictionary, and handing it to the translator. Meanwhile, using a *qualifier* only needs the original signature (and the prefix to be attached, which is always the name of the module): all the processing, including the creation of the qualified signature, the dictionary, and whatever else may be needed, is left to the qualifier.

At the end, we have implemented a class `Qualifier` which works precisely in that way. Its constructor creates the qualified version of the received signature and also the dictionary. The class has two public methods: `qTermStr` takes a string, assumed to contain a term, and returns the string representing the term with identifiers qualified. It does so by parsing the string in the original signature, building the equivalent term in the qualified one, and prettyprinting the result. Some terms need a special treatment, as discussed below. The other public method is `qId`, used to qualify isolated identifiers which are not terms, as when they appear in a declaration, or when a sort id appears in a membership axiom.

As a passing note, a `Qualifier` uses a single dictionary for all identifiers. This may cause problems in some extreme cases, because Maude allows a sort, an operator, and a variable to be declared with the same name while still being different objects. Using repeated names in that way would be silly, and, anyway, the same name would translate to the same qualified one, so probably there would be no problems after all. If problems do arise, the solution is as simple as keeping three separate dictionaries: for sorts, operators, and variables.

Names that are declared in imported modules are not qualified. Such modules are considered to be shared for some good reason (again, think of `BOOL` and `NAT`). We insist that we only used qualification on semi-flattened modules, but the qualifier does not check this. The qualifier qualifies all names declared in the text of the signature (which is given to it as a Maude module) and the inlined variables. There is only one exception: the operator `_@_` is not qualified. This operator is assumed to be named `_@_` all throughout, in atomic modules and in composed ones, in egalitarian and in standard ones. When a syncing criterion like `M$P = M'$P'` is translated into a condition for rules, the result uses that operator, with that name, unqualified. Therefore, the `_@_` is always left unqualified, even though it is declared in the module's signature. As a result, the name `_@_` appears in the text of the method `qTermStr`, which is quite ugly. I should try to improve this in the future. **TODO**

I considered (and implemented) not qualifying the names of operators that can be distinguished by their sorts. That is, given the declaration `op f : S -> S' .`, if sort `S` is declared in the module, and is therefore qualified, there is no need to qualify `f`, we can just use `op f : <M|S> -> ...`. No confussion can arise, because no operator from other module can take `<M|S>` as a parameter sort. This is nice in that it would consider the operator `_@_` as one case of a class of operators with *global names* guaranteed, the same in all modules in which they are used, which allows to view it as a less exceptional case. And, however, confusion does arise due to subsorting. Consider this example:
```maude
fmod A is
    sort SA .
endfm

mod B is
    pr A .
    sorts SB SB2 .
    subsort SA < SB .
    op f : SB -> SB2 .
endm

mod C is
    pr A .
    sorts SC SC2 .
    subsort SA < SC .
    op f : SC -> SC2 .
endm
```
After renaming as proposed above, we would leave `f` untouched and obtain:
```maude
subsort SA < <B|SB> .
op f : <B|SB> -> <B|SB2> .
```
and
```maude
subsort SA < <C|SC> .
op f : <C|SC> -> <C|SC2> .
```
But, because of subsorting, `<B|SB>` and `<C|SC>` are in the same kind, and the two `f`'s clash. As a result, we decided to qualify all declared names, with the known exception of `_@_`. **TODO**

I had similar doubts about the need to qualify inlined variables. Given that the scope of a variable like `V:Sort` is the equation or rule or membership axiom in which it appears, and that equally named appearances in different instructions are considered different, there seems to be no need for qualification. However, our specification of the Needham-Schoeder protocol gave us an example showing that it is indeed needed. In that example, both the initiator and the responder import the sort `Dict`, which is then declared to be a subsort of `State`. Then, we have in each system a rule involving an inlined variable named `D:Dict` (indeed, these are variables automatically inlined by our code, not inlined in the source cMaude code, but this makes no difference). After composition, this results in a rule in which two `D:Dict` are used, once from each module, and are considered the same, when they shouldn't. Thus, also all variables are qualified, either inlined or not.

There is more to be said about the technical means by which we perform the qualification of variables, `Qid`s, and numbers. The way to build a term with Rubén's tool is calling the method `makeTerm` for a `Symbol` object. `Symbol` is a class that represent Maude operators. The method `makeTerm` is passed a list of argument subterms, that is empty for constants. There are several methods to find the needed symbol, either from its name and argument kinds or from another term with that symbol, for example. For variables, however, their symbol is its sort (not really a sort, of course, but a particular `Symbol` object from which only the sort information can be drawn). That's the way Maude is implemented. But a variable term cannot be built from such a *pseudo-symbol*, so we need a different means. First, we check whether a term is a variable in this way: we create the string for a fake inlined variable with the same sort as the given `term`, `s = f'<fake|var>:{term.getSort()}'`; then, we parse said string in the given signature module to obtain a `Term`, `t = signatureModule.parseTerm(s)`; then, we compare the term's symbol with the one for the given term, `term.symbol() == t.symbol()`: if they are the same, the term is a variable with the given sort. When it happens to be a variable, we build the corresponding qualified variable term by using the method `parseTerm`, feeding it with a conveniently built string.

Similar considerations are needed for Maude builtin sorts. We have implemented the cases of `Qid` and `Nat`, but most probably others are needed (strings, for example). In the case of `Nat`, the symbol is always `s_`, and the list of arguments is always `[0]`, that is, just a zero. The reason, I guess, is that `s_` is an iterative operator (and maybe the problem in this case is not its builtin nature). There seems to be no way to find the number of iterations, but for a `Nat` we can just cast the term with `int`. Fortunately, using `parseTerm` also solves this problem. Of course, we do not qualify natural numbers, so the string to be given to `parseTerm` is easy to get.

Finally, let's discuss qualification of quoted identifiers, `Qid`'s. The first question is whether we should qualify them. On the one hand, they are used as identifiers *on the fly* and should be qualified as any other identifier. On the other hand, they may be seen as constants, not too different to natural numbers, and may be better left alone. We have decided not to qualify them, the main reason being that this is most convenient in our Needham-Schroeder protocol example, in which `Qid`'s are used for information interchange. Therefore, again the `parseTerm` method is used. To check if a term is a `Qid`, we check its symbol is stringified as `<Qids>`---that's what Maude produces.

(As a side note, Rubén's tool does not include the concept of session: all communication from Python to Maude is in a single session. Thus, if several `Qualifier` objects are used, their lifespans cannot overlap, or unexpected behaviours would happen. As Rubén says in the documentation, the package `multiprocessing` can help with that.)
