# On the definition of terminals using regex's

Author: Óscar Martín
Last modified: Apr 2021




In Lark, the parser I'm using for Python, terminals are described by regexes. When presented with a regular expression (describing a terminal symbol) and an input text, Lark chooses the largest initial segment of the input that matches the regular expression, and returns it as the terminal. The regexes needed to describe terminals for the Maude language turn out to be rather complex, mainly for two reasons. One reason is what a token is in Maude. A token is any sequence of non-whitespace characters, including puntuation ones, except that the seven *breaking characters* `()[]{},` are tokens by themselves, and they break a sequence of characters into several tokens if they are present, except if the are preceded by a backquote, that works as a escape. Thus, the sequence `ab(cd``[ef gh:ij` is composed of four tokens: `ab`, `(`, `cd``[ef`, and `gh:ij`. (Note that markdown requires the doubling of backquotes inside a backquoted string, but that before the bracket should be rendered as a simple one.)

The second reason Maude tokens are difficult to describe as regexes is that some particular tokens must not be accepted in certain places. For example, `=` is valid as a token. It is valid, for example, as the identifier for a user-defined operator. But on the lhs of an equation that cannot be accepted, because it can be confused with the `=` proper of the equation. Thus, there is not a unique concept of token, but an *equation lhs token*, a *rule lhs token*, etc., each with its list of exceptions.

Taking all this into account results in really long regexes, impossible to parse for the human eye. That's why I have used regex generating functions in my Python code. I describe them in this document.

The main function I use is called `token`. Its header reads
```py
def token(*ss: str) -> str:
```
It receives a sequence of tokens, and returns a regex that matches any of those tokens, but only when they appear as such, that is, surrounded by whitespace or non-backquoted breaking chars. It is assumed that each argument can be a token, that is, that it does not contain any whitespace, and non-backquoted breaking characters can only appear by themselves. Funny things may happen otherwise. Whtn the sequence of arguments is empty, the produced regex matches any token.

The regex for a general token is obtained like this:
```py
rf'((?<!`)[{BRCHARS}])'\
+ '|'\
+ '('\
+ rf'(^|(?<=\s|(?<!`)[{BRCHARS}]))'\
+ rf'((([^\s{BRCHARS}]|(?<=`)[{BRCHARS}]))+)'\
+ rf'(?=\s|[{BRCHARS}]|$)'\
+ ')'
```
The constant `BRCHARS` is defined to be `BRCHARS = r'\[\]\{\}\(\),'`. The main operator in the regex is the disjunction in line 2. The first disjunct, `(?<!``)[{BRCHARS}]`, matches any breaking char not preceded by a backquote. The second disjunct consists, according to line 5, `([^\s{BRCHARS}]|(?<=``)[{BRCHARS}])+`, in a sequence of one or more characters that either are not whitespace nor breaking chars, or are breaking chars preceded by a backquote. But this, according to line 4, `^|(?<=\s|(?<!``)[{BRCHARS}])`, must to be preceded by, either the beginning of the input, or whitespace, or a non-backquoted breaking char. And, according to line 6, `?=\s|[{BRCHARS}]|$`, it must be followed by whitespace or a breaking char.

In all regexes I generate, I prefer to add as many parentheses as possibly needed, because the resulting expressions may be used as part of other larger ones, and may interact in unexpected ways if no parentheses are present.

The previous regex is returned when the sequence of arguments is empty. Otherwise, a disjunction is produced, with a disjunct to match each argument token. If the argument is a breaking char, a lookbehind is added to check it is not preceded by a backquote. Otherwise, the same lookarounds are added as in lines 4 and 6 above. I insist that, in this case, we do not check that the argument is a valid token.

This is for matching tokens. What about the exceptions? That is, how do we produce a regex that matches any token but a few given ones? This function is the answer:
```py
def anyTokenBut(*ss: str) -> str:
    return rf'(?!{token(*ss)})({token()})'
```

Identifiers and tokens in Maude are, in principle, two different kinds of terminals. But, to the best of my knowledge, their definitions as sequences of characters are the same. The exceptions in this case are that a single dot must not be allowed as identifier, and `<` must not be allowed as a sort identifier. To remind the reader (and the writer): identifiers are what appear in declarations of sorts, operators and variables (and can appear also elsewhere, or course); tokens are any of the pieces that for a term. An identifier can be a token, but also other things, like the parentheses needed to enclose the arguments of an operator. Pieces of identifiers can also be tokens, as for mixfix declared operators: `op _[_] : ...` and then the term `a[b]` contains four tokens, which can be parsed to three identifiers. 

Sequences of tokens can be enclosed in parentheses, and the result is again called a token in Maude. An obvious use of parentheses is to enclose the comma-separated sequence of arguments to a function. Another important use is to disambiguate some expressions, or make them ready for the Maude interpreter. For example, if we really need a `=` on the lhs lhs of an equation, we can do it: `eq (a = b) = c .`

Strictly speaking, a parenthesized token is a sequence of tokens enclosed in parentheses. That's how it is defined in Maude's grammar (Chapter 24 in the manual). However, I do not need to recursively parse inside the parentheses: I can just accept as a term whatever is enclosed in a pair of balanced parentheses, which is quite cheaper.

Matching balanced parentheses is well-known to be outside the capabilities of regexes. (Side note: some extensions of regexes have this capability. For example, the Python package  `regex` (instead of the more standard `re`). But Lark does not support this new package, or rather, only very partially.) I found [here](https://stackoverflow.com/questions/5454322/python-how-to-match-nested-parentheses-with-regex/22809985#22809985) a way to generate regexes that match balanced parentheses up to some given level. I have adapted it like this:
```py
def balParens(n: int):
    return r'\(' + r'[^()]*(\(' * n + r'[^()]*' + r'\)[^()]*)*' * n + r'\)'
```
This produces a regex that matches any string with balanced parentheses up to the nested level given by the argument (plus the most external ones). The resulting regex is not perfect: it does not take into account backquoted parentheses. Moreover, individual non-backquoted parentheses can be used in Maude as identifiers, and be used anywhere, producing unbalanced expressions, but theoretically valid ones.

With this, we can define terms (that are not terminals in our grammar). There are different kinds of terms, considering the tokens that cannot be part of them (unless parenthesized). For example, the definition for the term on the lhs of an equation is given by this Python f-string, to be given, once evaluated, to Lark as part of the grammar:
```
TOKEN_PAREN: /{balParens(5)}/
TOKEN_EQ_LHS: /{anyTokenBut('=')}/
term___eq_lhs: (TOKEN_PAREN | TOKEN_EQ_LHS)+
```
That is, we accept balanced parentheses up to the fifth level (plus the most external ones), which should be enough for normal uses. And a term is a sequence of parenthesized and not parenthesized tokens, with the exception, in the latter case, of the equals sign.

In the rest of this document I comment on some particular cases.

The definition of the tokens for the lhs of a condition deserve some attention. First, because it has to take as exceptions the symbols for all possible conditions, that is, `=>`, `:=`, `:`, and `=`. (This can be slightly refined, because rewrite conditions cannot appear in equations nor in membership axioms, but I'm using the same regex for all condition tokens.) This is in contrast with tokens for equations, which only need to avoid `=`, for rules, which only need to avoid `=>`, and for membership axioms, that only need to avoid `:`. The reason is that each type of instruction is introduced by a different keyword, but all conditions look the same at the beginning.

Quite the same happens with rules, when egalitarian ones are taken into account. Both types of rules are introduced by `rl`, but their lhs ends with different symbols, `=>` or `=[`. So both exceptions are included in the definition of `TOKEN_RL_LHS`.

Again about conditions, a case of particular interest is the one for Boolean conditions, that is, `... if C .`, where `C` is a Boolean expression. This is accepted in Maude as a convenient shortcut for `... if C = true .` But the term `C` is not restricted in what it can contain. Thus, there is always the possibility of an ambiguity between, for example, `if C1 := C2` being parsed as an equational condition or as a shortcut for `if (C1 := C2) = true`. The way Maude parses may allow it to determine whether `C1 := C2` is a valid term in the module's signature, but not for me, because my parsing of terms is performed separatedly. (See the document about parsing in two passes.) (By the way, I don't know how Full Maude deals with this.)

A possible solution would have been to ask the coder to write parentheses around any expression that could be ambiguous in this sense. I think this would have been reasonable. But I have found another solution in a feature of Lark that is not part of standard EBNF: priorities. All grammatical symbols have a priority of 1 by default in Lark, but higher or lower (even negative) priorities can be given to particular symbols in their definitions. In our case: `bool_condition.-1: ...` allows a condition to be parsed as a Boolean one only if no other parsing is possible. In order to keep our grammar definition as standard as possible, we only use priorities in this and another one case.

Some of the exceptions we consider when defining tokens may not be obvious. For example, tokens appearing on the lhs of a condition need the single dot as an exception. This is odd, because the lhs of a condition never ends in a dot: this is a problem for the rhs. But, were the dot not included as an exception, two consecutive instructions like `ceq a = b if c = d . eq a' = b' .` could be parsed as `ceq a = b if (c = d . eq a') = b' .`

I have not been able to produce regexes that avoid all surrounding whitespace, so I always `strip` the result of `str`ing each `Token`. This is only evident in Python code, not in the grammar.

The terminals called `_SORT_LIT` and similar ones deserve a few lines. The definition of `_SORT_LIT` is the regex `sorts?`. That is, either `sort` or `sorts` (which are synonyms in Maude). But if I include directly `sorts?` in a rule, like
```lark
sort_decl: /sorts?/ id_list "."
```
(slashes delimit regexes in Lark), directly in a rule, it produces a 
token in the AST (which kind of makes sense: because it can match either 
`sort` or `sorts`, it produces a token to inform of which is the case). 
Giving to terminals names starting with an underscore, like `_SORT_LIT`, 
tells Lark to avoid the token. By the way, for some reason I do not know, if I write instead `"sorts" | "sort"` sometimes the `s` is 
included (if present) and sometimes it isn't, in a seemingly non-deterministic way. The regex solution works well.
