# Parsing Maude code---a tale of two passes

Author: Óscar Martín
Last modified: Apr 2021



An important ingredient in the desing of the Maude language is that it allows a very flexible *mixfix* syntax for operators. This is reflected in the use of underscores as placeholders in declarations, and also in the fact that any symbol can be part of an identifier. This, certainly, comes at a price. One consequence is the need for parsing in two passes: the first pass delimits the terms and other syntactical constructs according to Maude's own syntax; the second pass parses into terms using the syntax defined in the module's declarations. For example, when confronted with `eq a @ f(b) = g(a + b) .`, the first pass, based on Maude' syntax definition, is able to identify five pieces: `eq`, `a @ f(b)`, `=`, `g(a + b)`, and `.`, but it is not able to go into the two terms `a @ f(b)` and `g(a + b)`, because their structure depends on its particular module signature. The first pass also parses the declarations of sorts, subsorts, and operators. The syntax of these declarations only depend on Maude's syntax. For example, `sort S . op a : -> S .` declares a sort named `S` and a constant `a` with that sort. Sort, subsort and operator declarations are the signature on which the terms appearing in the rest of the module are then parsed in the second pass.

There are two ways in which two-pass parsing can be performed. In the first way, the first pass analyses all the text in the module and delimits and isolates (in so-called *bubbles*) all terms in which it is not able to parse. As a result of the first pass, the local grammar for the module is obtained, and this is then used to parse into the bubbles (to *solve* them) using Maude's `metaParse` function (implemented in C++ and declared in Maude's prelude with a hook). This seems to be the way Full Maude works.

In the second way, the first pass only looks for sort, subsort and operator declarations. With them, it creates a local grammar that is joined to Maude's grammar and the whole module is parsed against this complete global+local grammar. This seems to be the way Maude works. (I have not looked into Maude's implementation, so this is just what I deduce from its workings.) 

I decided to use a different parser for each phase (Lark for the first, `metaParse` for the second). Each parser uses a different means for describing syntax (EBNF for Lark, Maude modules for `metaParse`). Thus, I have to work in the first way. See more on my choices below.

As a particular example of the how flexible Maude's syntax is, how this can be problematic, and how two-pass parsing is used, consider this piece of code:
```maude
fmod KK is
    sort S .
    op owise : -> S .
    op `[_`] : S -> S .
    op __ : S S -> S .
    eq [owise] = [owise] [owise] .
endfm
```
The equation is syntactically valid (although nonterminating). Maude warns it is ambiguous, which it can only know if it parses it against the composed global+local grammar: Maude's  global grammar is needed to identify `[owise]` as an attribute list; the local module's grammar is needed to identify it as a term of sort `S`. Full Maude, on the other hand, accepts the equation and (as I deduce from other examples) seems to interpret the rightmost `[owise]` as an attribute list. I see, looking into its code, that Full Maude is doing something odd: the first pass parses everything to the right of the `=` of an `eq` into a bubble, that is then processed in code, as if *manually parsing* it into a term and an optional attribute list.

Another example that shows the two different ways is with conditional equations like this:
```maude
ceq a = if b then c else d fi if e .
```
While Maude accepts it and interprets it correctly, Full Maude complains when it arrives at the first `if`. So does the cMaude tool. The simple solution, for both Full Maude and cMaude, is enclosing the `if...fi` part in parenthesis.

It seems to me that warning about ambiguities (as Maude does) is the best way. But I cannot do that, because cMaude's two passes are clearly separated by the use of two different parsers. cMaude, when in doubt, gives priority to the attribute-list interpretation over the term one. So it works like Full Maude in this respect.

It is not impossible that, in a future time, I use Maude's `metaParse` as well for the first pass, maybe as an alternative to Lark. If this happens, I would be able to perform the two passes the Maude way. I have also tried to translate Maude's syntax for module signatures into EBNF form to use Lark also in the second pass. This roughly works, but such translation is not a straightforward process, as I discuss somewhere else.
