# Using `metaParse` from Python

Author: Óscar Martín
Last modified: Apr 2021



It is explained in [another document](lark%20and%20metaparse.md) that parsing using Maude tools requires working at the metalevel, specially (only?) to be able to define and use terminals. Also in that document is explained how to invoke Maude functions from Python using Rubén Rubio's bindings, included in his Python package `maude`. This document explains the precise way to use these tools to parse code from Python using Maude's `metaParse`.

So, yes, the Maude function we need to use is `metaParse`. Rubén's tool does not include a method that directly emulates `metaParse` (for good reasons, no doubt). Instead, the way to use `metaParse` is (i) writing a function call we need, making a string of it; (ii) parsing that string using Rubén's method `parseTerm`, and (iii) calling `reduce` on the resulting term. In our case:
```py
term = myParserMod.parseTerm('metaParse(WHOLE-GRAMMAR-META, tokens, type)')
term.reduce()
```
What I call `myParserMod` is the module (precisely, an object in Rubén's class `Module`---we see below how to obtain it) on which the `metaParse`-rooted term is going to be parsed. Thus, everything in that term must make sense in `myParserMod`: `metaParse`, `WHOLE-GRAMMAR-META`, `tokens`, and `type`:
- The operator `metaParse` is declared in the module `META-LEVEL`, in Maude's prelude. So `META-LEVEL` must be imported into `wholeGrammar`.
- What I have called `WHOLE-GRAMMAR-META` is the name of a constant of variable that contains the metarepresentation of a module (precisely, of a term of sort `FModule`, as declared in Maude's prelude), usually consisting just of a signature (i. e., just declarations) on which the sequence of `tokens` is to be parsed. Such `FModule` could be written literally inline, but it is certainly more usual to give it a name and use it.
- `tokens` is a sequence of tokens supposedly representing a term valid in the grammar described in `WHOLE-GRAMMAR-META`. Each token is represented as a `Qid`, so the module `QID` has to be imported into `myParseMod`. (Indeed, `META-LEVEL` imports `QID`, so an explicit importation is not strictly needed.)
- Finally, `type` is a `Qid` with the name of a sort or kind to which supposedly the term given by the `tokens` can be parsed to. The particular value `anyType` can be used here.

The reason I insist in adding *whole* to some names is to make it clear the inclusion of both terminals and the proper grammar. This is made clear below.

Note the two different parses that are invoked at once in the sentence `myParserMod.parseTerm('metaParse(WHOLE-GRAMMAR-META, tokens, type)')`. The inner one is the one we are interested in: parsing `tokens` in the grammar defined in `WHOLE-GRAMMAR-META`. But `metaParse` is a Maude function, and we can only call Maude functions with the help of `parseTerm` and `reduce`. At the end, however, it is Maude's internal parser which makes all the parsing work, in two different ways.

All in all, the module represented by the variable `myParseMod` used above has this form:
```py
preMyParser = """
    fmod MY-PARSER is
        including META-LEVEL .
        protecting QID .
        op WHOLE-GRAMMAR-META : -> FModule .
        eq WHOLE-GRAMMAR-META = (
            fmod 'WHOLE-GRAMMAR is
                < grammar definition >
            endfm
        ) .
    endfm
"""
```
I have called it `pre...` because some processing, explained below, must be performed on it before use. The result of such processing produces the string `myParser`, which must be feed into Maude with
```py
maude.input(myParser)
```
Then,
```py
myParserMod = maude.getModule("MY-PARSER")
```
retrieves in `myParseMod` the `Module` object on which the `metaParse` term can be parsed and reduced as given above:
```py
term = myParserMod.parseTerm('metaParse(WHOLE-GRAMMAR-META, tokens, type)')
term.reduce()
```

Two points remain to be clarified: what goes inside the Maude module I have called `WHOLE-GRAMMAR`, and how the sequence of `tokens` can be obtained from a string, which is the input we are able to produce.

Let's start with the second one. This is it:
```py
def tokenize(expr: str) -> str:
    m = maude.getModule('LEXICAL')
    t = m.parseTerm(f'tokenize("{expr}")')
    t.reduce()
    return str(t)
```
Again, we use the trick of parsing and reducing a term containing the Maude call we want to execute. In this case, Maude's `tokenize` function, defined in the module `LEXICAL`, in Maude's prelude. When fed with the string `"sort S . op []_ : S -> S ."`, it produces the string `"'sort 'S '. 'op '``[ '``] '_ ': 'S '-> 'S '."` (As always, double backquotes are required by markdown, but must be simple ones in Maude code.) This is still a string for Python.

The final point to clarify is about what must go inside the Maude module I have named `WHOLE-GRAMMAR` and stored under the constant `WHOLE-GRAMMAR-META`. Remember: this is the (metarepresented) signature on which the `tokens` are going to be parsed into a term. It can be seen as consisting in two parts: the definition of *tokens*, and the rest of the grammar. These are quite different tasks. The latter needs to be done directly at the metalevel; the former preferibly not. All this follows the way Full Maude does it.

There is some room for confusion with the use of the work *token*. Sometimes it is used as mostly any `Qid` (with the known exceptions about breaking characters). This is what the function `tokenize` produces. Now, however, tokens represent the leaves of a parse tree. This new kind of token is built with a constructor operator like `sortToken` that takes any `Qid` as its argument (usually with some exceptions). So, at the end, they are not very different things: `'S` is a token in the first sense; `sortToken('S)` is a token in the second sense.

Tokens (in the new sense) are declared in Maude by means of hooks. That is, they are ultimately defined in C++ code. For example, this is the operator declaration of sort tokens (that is, tokens that can be used as the name of a sort), taken from Full Maude:
```maude
op 'sortToken : 'Qid -> 'SortToken
    [special(
        (id-hook('Bubble, '1 '1)
         op-hook('qidSymbol, '<Qids>, nil, 'Qid)
         id-hook('Exclude, '`[ '`] '< 'to '`, '. '`( '`) '`{ '`} ':
                           '| 'ditto 'precedence 'prec 'gather
                           'assoc 'associative 'comm 'commutative
                           'ctor 'constructor 'id: 'strat 'strategy
                           'poly 'memo 'memoization 'iter 'frozen
                           'config 'object 'msg 'metadata 'nonexec
                           'variant)))] .
```
This is metarepresented. It seems there is no other way. Usually, the way to get the metalevel representation of anything, from a term to a whole module, is writing it in plain Maude and then using Maude-provided functions to perform the metarepresentation. In this case, however, it is not possible. Indeed, it seems the function `upModule` does not take `special` attributes into account. So all needed kinds of tokens must be defined in this way. Fortunately, the hook mechanism is powerful enough. Full Maude has definitions for all needed token types, and more can be defined in the same way if needed.

For some reason I'm not sure I understand, the module `QID-LIST` needs to be imported into the module where token operators are defined. Such importation is easily metarepresented as `including 'QID-LIST .`.

Finally, for `WHOLE-GRAMMAR` to deserve its name, the rest of the grammar must be included in it. It is possible to write it directly at the metalevel, as with the tokens. Possible but impractical. The way we do it, following Full Maude, is to write the module in plain Maude and `including` it in `WHOLE-GRAMMAR`. For example, suppose we want to define the grammar for a set of sort declarations:
```maude
fmod MY-GRAMMAR is
    sorts SortDecl SortToken .
    op sort_. : SortToken -> SortDecl .
    op __ : SortDecl SortDecl -> SortDecl .
endfm
```
We have to feed this to Maude and add `including 'MY-GRAMMAR .` to the `WHOLE-GRAMMAR` module. Thus, when Maude needs to use `WHOLE-GRAMMAR` it already has `MY-GRAMMAR` available, and the importation proceeds. This way allows us to write `MY-GRAMMAR` at base level. For `MY-GRAMMAR` to be accepted by Maude, `'SortToken` has to be declared in it. Then, it is used both in `MY-GRAMMAR` and in the definition of tokens.

We may be interested in working with several grammars based on the same tokens, so it seems better to keep both parts separated and allow for some form of *parameterization*. What we do is the following: in `WHOLE-GRAMMAR` we insert the instruction `including '<youR|grammaR|fmoD|namE|herE> .` The funny name is a placeholder. Then we execute
```py
myParser = preMyParser.replace("<youR|grammaR|fmoD|namE|herE>", 'MY-GRAMMAR')
```
changing `MY-GRAMMAR` for whatever other name. This is the preprocessing we talked about some paragraphs above.

Thus, the whole module `MY-PARSER` is like this:
```py
preMyParser = """
    fmod MY-PARSER is
        including META-LEVEL .
        op WHOLE-GRAMMAR-META : -> FModule .
        eq WHOLE-GRAMMAR-META = (
            fmod 'WHOLE-GRAMMAR is
                including '<youR|grammaR|fmoD|namE|herE> .
                including 'QID-LIST .
                sorts none .
                none
                op 'sortToken : 'Qid -> 'SortToken
                    [special(
                    (id-hook('Bubble, '1 '1)
                    op-hook('qidSymbol, '<Qids>, nil, 'Qid)
                    id-hook('Exclude, '`[ '`] '< 'to '`, '. '`( '`) '`{ '`} ':
                                        '| 'ditto 'precedence 'prec 'gather
                                        'assoc 'associative 'comm 'commutative
                                        'ctor 'constructor 'id: 'strat 'strategy
                                        'poly 'memo 'memoization 'iter 'frozen
                                        'config 'object 'msg 'metadata 'nonexec
                                        'variant)))] .
                none
                none
            endfm
        ) .
    endfm
"""
```

With all this considerations, I have coded this Python function:
```py
def parseIn(expr: str, grm: str, tp: Optional[str] = None) -> maude.Term:
    maude.input(myGrammar)
    maude.input(grm)
    tp = 'anyType' if tp is None else "'" + tp
    m = maude.getModule("MY-PARSER")
    term = m.parseTerm(
        f"metaParse(WHOLE-GRAMMAR-META, {tokenize(expr)}, {tp})")
    term.reduce()
    return term
```
It receives the string to be parsed, the name of the grammar module to parse it in, and, optionally, the type. It returns the result of the *metaparsing*. For example,
```py
maude.init(advise=False)
myParser = preMyParser.replace("<youR|grammaR|fmoD|namE|herE>", 'MY-GRAMMAR')
t = parseIn("sort S . sort T .", myParser, "SortDecl")
print(t)
t = parseIn("sort TTT .", myParser)
print(t)
```
prints
```
{'__['sort_.['sortToken[''S.Qid]],'sort_.['sortToken[''T.Qid]]],'SortDecl}
{'sort_.['sortToken[''TTT.Qid]],'SortDecl}
```

For reference, this complete Python script works:
```py
import maude
from typing import Optional


myGrammar = """
fmod MY-GRAMMAR is
    sorts SortDecl SortToken .
    op sort_. : SortToken -> SortDecl .
    op __ : SortDecl SortDecl -> SortDecl .
endfm
"""


preMyParser = """
fmod MY-PARSER is
    including META-LEVEL .
    op WHOLE-GRAMMAR-META : -> FModule .
    eq WHOLE-GRAMMAR-META = (
        fmod 'WHOLE-GRAMMAR is
            including '<youR|grammaR|fmoD|namE|herE> .
            including 'QID-LIST .
            sorts none .
            none
            op 'sortToken : 'Qid -> 'SortToken
                [special(
                  (id-hook('Bubble, '1 '1)
                   op-hook('qidSymbol, '<Qids>, nil, 'Qid)
                   id-hook('Exclude, '`[ '`] '< 'to '`, '. '`( '`) '`{ '`} ':
                                     '| 'ditto 'precedence 'prec 'gather
                                     'assoc 'associative 'comm 'commutative
                                     'ctor 'constructor 'id: 'strat 'strategy
                                     'poly 'memo 'memoization 'iter 'frozen
                                     'config 'object 'msg 'metadata 'nonexec
                                     'variant)))] .
            none
            none
        endfm
    ) .
endfm
"""


def tokenize(expr: str) -> str:
    m = maude.getModule('LEXICAL')
    t = m.parseTerm(f'tokenize("{expr}")')
    t.reduce()
    return str(t)


def parseIn(expr: str, grm: str, tp: Optional[str] = None) -> maude.Term:
    maude.input(myGrammar)
    maude.input(grm)
    tp = 'anyType' if tp is None else "'" + tp
    m = maude.getModule("MY-PARSER")
    term = m.parseTerm(f"metaParse(WHOLE-GRAMMAR-META, {tokenize(expr)}, {tp})")
    term.reduce()
    return term


maude.init(advise=False)
myParser = preMyParser.replace("<youR|grammaR|fmoD|namE|herE>", 'MY-GRAMMAR')
t = parseIn("sort S . sort T .", myParser, "SortDecl")
print(t)
t = parseIn("sort TTT .", myParser)
print(t)
```
